unit DACOVARS; {Unit 7 of CONTEXT.PAS}

interface

  uses
    GLOBALS, MATHS, UTILS;

  procedure ARMACovariances (var ARMASpec: ARMARecord;
                  var gamma: shortVector;
                  lags: integer);

  procedure ARMACorrelations (lags: integer;
                  var rho: shortVector;
                  var gamma: shortVector;
                  var autoMin: real);

  function DPartialNumerator (var beta: shortVector;
                  var lambda: complex;
                  var q: integer): complex;

  procedure DPartialDenoms (var mu: complexVector;
                  var denominator: complex;
                  p, j: integer);

  procedure DPartialCoefficients (var partialC: complexVector;
                  var ARMASpec: ARMARecord);

  procedure DGammaCoeffs (var gCoeff: complexVector;
                  var ARMASpec: ARMARecord);


implementation

{_________________________________________________}

  procedure ARMACovariances (var ARMASpec: ARMARecord;
                  var gamma: shortVector;
                  lags: integer);

{This is the procedure for finding a sequence of ARMA autocovariances}
{equal in number to lags}

    var
      i, j, t, tau, r: integer;
      delta, f: shortVector;
      a: matrix;

  begin {Find the delta values}

    with ARMASpec do
      begin {with ARMASpec}
        r := Max(p, q);
        for t := 0 to q do
          begin
            delta[t] := beta[t] * varEpsilon;
            tau := Min(t, p);
            for i := 1 to tau do
              delta[t] := delta[t] - delta[t - i] * alpha[i];
            delta[t] := delta[t] / alpha[0]
          end;

        for i := 0 to r do
          for j := 0 to r do
            begin {i,j : form the matrix of alpha values}
              a[i, j] := 0.0;
              if ((i - j) >= 0) and ((i - j) <= p) then
                a[i, j] := alpha[i - j];
              if ((i + j) <= p) and (j > 0) then
                a[i, j] := a[i, j] + alpha[i + j];
            end; {i, j}

        for i := 0 to r do
          begin{i : form the RHS vector}
            f[i] := 0.0;
            for j := i to q do
              f[i] := f[i] + beta[j] * delta[j - i]
          end; {i}

{Solve for the initial autocovariances}
        LUsolve(0, r + 1, a, gamma, f);

{Find the succeeding autocovariances}
        for i := r + 1 to lags do
          begin {i}
            gamma[i] := 0.0;
            for j := 1 to p do
              gamma[i] := gamma[i] - alpha[j] * gamma[i - j];
            gamma[i] := gamma[i] / alpha[0];
          end; {i}

      end; {with ARMASpec}

    for i := 0 to 10 do
      ARMASpec.aCovar[i] := gamma[i];
    ARMASpec.aCovarsFound := true;
{Why do we need a separate gamma vector?}

  end; {ARMACovariances}

{_________________________________________________}

  procedure ARMACorrelations (lags: integer;
                  var rho: shortVector;
                  var gamma: shortVector;
                  var autoMin: real);

    var
      i: integer;
  begin
    autoMin := 0.0;
    for i := 1 to lags do
      begin
        rho[i] := gamma[i] / gamma[0];
        if rho[i] < autoMin then
          autoMin := rho[i];
      end;
    autoMin := Abs(autoMin);
    rho[0] := 1;

  end; {ARMACorrelations}

{_________________________________________________}

  function DPartialNumerator (var beta: shortVector;
                  var lambda: complex; {lambda is the inverse mu}
                  var q: integer): complex;

{This procedure evaluates the numerator of beta(z)/alpha(z) at z = lambda}

    var
      polyVal: complex;
      i: integer;

  begin
    polyVal.re := beta[q];
    polyVal.im := 0.0;
    for i := 1 to q do
      begin
        ComplexMult(polyVal, lambda, polyVal);
        polyVal.re := polyVal.re + beta[q - i];
      end;
    DPartialNumerator := polyVal;
  end; {DPartialNumerator}

{_________________________________________________}

  procedure DPartialDenoms (var mu: complexVector;
                  var denominator: complex;
                  p, j: integer);

{This procedure evaluates the denominator of beta(z)/alpha(z) at z = lambda}

    var
      i: integer;
      factor, lambda: complex;

  begin
    ComplexInv(mu[j], lambda);
    denominator.re := 1.0;
    denominator.im := 0.0;
    for i := 1 to p do
      begin {i}
        if i <> j then
          begin
            ComplexMult(mu[i], lambda, factor);
            factor.re := 1.0 - factor.re;
            factor.im := -factor.im;
            ComplexMult(denominator, factor, denominator);
          end;
      end; {i}

  end; {PartialDenoms}

{_________________________________________________}

  procedure DPartialCoefficients (var partialC: complexVector;
                  var ARMASpec: ARMARecord);
    var
      beta: shortVector;
      mu: complexVector;
      lambda, denominator, numerator: complex;
      p, q, j: integer;

  begin
    p := ARMASpec.p;
    q := ARMASpec.q;
    mu := ARMASpec.poles;
    beta := ARMASpec.beta;

    for j := 1 to p do
      begin
        DPartialDenoms(mu, denominator, p, j);
        ComplexInv(mu[j], lambda);
        numerator := DPartialNumerator(beta, lambda, q);
        ComplexDiv(numerator, denominator, partialC[j]);
      end;
  end; {PartialCoefficients;}

{_________________________________________________}

  procedure DGammaCoeffs (var gCoeff: complexVector;
                  var ARMASpec: ARMARecord);

    var
      i, j, p: integer;
      numerator, denominator, ratio: complex;
      mu, partialC: complexVector;

  begin
    p := ARMASpec.p;
    mu := ARMASpec.poles;

    DPartialCoefficients(partialC, ARMASpec);

    for i := 1 to p do
      begin {i}
        gCoeff[i].re := 0.0;
        gCoeff[i].im := 0.0;
        for j := 1 to p do
          begin {j}
            ComplexMult(partialC[i], partialC[j], numerator);
            ComplexMult(mu[i], mu[j], denominator);
            denominator.re := 1 - denominator.re;
            denominator.im := -denominator.im;
            ComplexDiv(numerator, denominator, ratio);
            ComplexAdd(gCoeff[i], ratio, gCoeff[i]);
          end; {j}
      end; {i}

  end; {GammaCoeffs}

{_________________________________________________}

end. {DAC0VARS:  Unit 9 of CONTEXT.PAS}

